@@ -19,15 +19,11 @@
 #ifndef vte_vte_h_included
 #define vte_vte_h_included
 
-#ident "$Id: vte.h,v 1.66 2003/05/19 20:55:10 nalin Exp $"
-
-#include <sys/types.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <X11/Xlib.h>
 #include <glib.h>
-#include <pango/pango.h>
 #include <gtk/gtk.h>
+#include <pango/pango.h>
+
+#include <sys/types.h> /* for pid_t */
 
 G_BEGIN_DECLS
 
@@ -63,6 +59,47 @@ struct _VteTerminalClass {
 	/* Inherited parent class. */
 	GtkWidgetClass parent_class;
 
+	/*< protected > */
+	/* Default signal handlers. */
+	void (*eof)(VteTerminal* terminal);
+	void (*child_exited)(VteTerminal* terminal);
+	void (*emulation_changed)(VteTerminal* terminal);
+	void (*encoding_changed)(VteTerminal* terminal);
+	void (*char_size_changed)(VteTerminal* terminal, guint char_width, guint char_height);
+	void (*window_title_changed)(VteTerminal* terminal);
+	void (*icon_title_changed)(VteTerminal* terminal);
+	void (*selection_changed)(VteTerminal* terminal);
+	void (*contents_changed)(VteTerminal* terminal);
+	void (*cursor_moved)(VteTerminal* terminal);
+	void (*status_line_changed)(VteTerminal* terminal);
+	void (*commit)(VteTerminal* terminal, gchar *text, guint size);
+
+	void (*deiconify_window)(VteTerminal* terminal);
+	void (*iconify_window)(VteTerminal* terminal);
+	void (*raise_window)(VteTerminal* terminal);
+	void (*lower_window)(VteTerminal* terminal);
+	void (*refresh_window)(VteTerminal* terminal);
+	void (*restore_window)(VteTerminal* terminal);
+	void (*maximize_window)(VteTerminal* terminal);
+	void (*resize_window)(VteTerminal* terminal, guint width, guint height);
+	void (*move_window)(VteTerminal* terminal, guint x, guint y);
+
+	void (*increase_font_size)(VteTerminal* terminal);
+	void (*decrease_font_size)(VteTerminal* terminal);
+
+	void (*text_modified)(VteTerminal* terminal);
+	void (*text_inserted)(VteTerminal* terminal);
+	void (*text_deleted)(VteTerminal* terminal);
+	void (*text_scrolled)(VteTerminal* terminal, gint delta);
+	void (*copy_clipboard)(VteTerminal* terminal);
+	void (*paste_clipboard)(VteTerminal* terminal);
+
+	/* Padding for future expansion. */
+	void (*vte_reserved1)(void);
+	void (*vte_reserved2)(void);
+	void (*vte_reserved3)(void);
+	void (*vte_reserved4)(void);
+
 	/*< private > */
 	/* Signals we might emit. */
 	guint eof_signal;
@@ -113,17 +150,32 @@ typedef enum {
 	VTE_ERASE_DELETE_SEQUENCE
 } VteTerminalEraseBinding;
 
+/* Values for the anti alias setting */
+typedef enum {
+	VTE_ANTI_ALIAS_USE_DEFAULT,
+	VTE_ANTI_ALIAS_FORCE_ENABLE,
+	VTE_ANTI_ALIAS_FORCE_DISABLE
+} VteTerminalAntiAlias;
+
 /* The structure we return as the supplemental attributes for strings. */
 struct _VteCharAttributes {
 	long row, column;
 	GdkColor fore, back;
-	gboolean underline:1, strikethrough:1;
+	guint underline:1, strikethrough:1;
 };
 typedef struct _VteCharAttributes VteCharAttributes;
 
+/* The name of the same structure in the 0.10 series, for API compatibility. */
+struct vte_char_attributes {
+	long row, column;
+	GdkColor fore, back;
+	guint underline:1, strikethrough:1;
+};
+
 /* The widget's type. */
 GtkType vte_terminal_get_type(void);
 GtkType vte_terminal_erase_binding_get_type(void);
+GtkType vte_terminal_anti_alias_get_type(void);
 
 #define VTE_TYPE_TERMINAL		(vte_terminal_get_type())
 #define VTE_TERMINAL(obj)		(GTK_CHECK_CAST((obj),\
@@ -141,6 +193,9 @@ GtkType vte_terminal_erase_binding_get_type(void);
 #define VTE_TYPE_TERMINAL_ERASE_BINDING	(vte_terminal_erase_binding_get_type())
 #define VTE_IS_TERMINAL_ERASE_BINDING(obj)	GTK_CHECK_TYPE((obj),\
 						VTE_TYPE_TERMINAL_ERASE_BINDING)
+#define VTE_TYPE_TERMINAL_ANTI_ALIAS	(vte_terminal_anti_alias_get_type())
+#define VTE_IS_TERMINAL_ANTI_ALIAS(obj)		GTK_CHECK_TYPE((obj),\
+						VTE_TYPE_TERMINAL_ANTI_ALIAS)
 
 /* You can get by with just these two functions. */
 GtkWidget *vte_terminal_new(void);
@@ -151,11 +206,18 @@ pid_t vte_terminal_fork_command(VteTerminal *terminal,
 				gboolean utmp,
 				gboolean wtmp);
 
+/* Users of libzvt may find this useful. */
+pid_t vte_terminal_forkpty(VteTerminal *terminal,
+			   char **envv, const char *directory,
+			   gboolean lastlog,
+			   gboolean utmp,
+			   gboolean wtmp);
+
 /* Send data to the terminal to display, or to the terminal's forked command
  * to handle in some way.  If it's 'cat', they should be the same. */
 void vte_terminal_feed(VteTerminal *terminal, const char *data, glong length);
-void vte_terminal_feed_child(VteTerminal *terminal,
-			     const char *data, glong length);
+void vte_terminal_feed_child(VteTerminal *terminal, const char *text, glong length);
+void vte_terminal_feed_child_binary(VteTerminal *terminal, const char *data, glong length);
 
 /* Copy currently-selected text to the clipboard, or from the clipboard to
  * the terminal. */
@@ -164,11 +226,15 @@ void vte_terminal_paste_clipboard(VteTerminal *terminal);
 void vte_terminal_copy_primary(VteTerminal *terminal);
 void vte_terminal_paste_primary(VteTerminal *terminal);
 
+/* simple manipulation of selection */
+void vte_terminal_select_all(VteTerminal *terminal);
+void vte_terminal_select_none(VteTerminal *terminal);
+
 /* Set the terminal's size. */
 void vte_terminal_set_size(VteTerminal *terminal,
 			   glong columns, glong rows);
 
-/* Set various one-off settings. */
+/* Set various on-off settings. */
 void vte_terminal_set_audible_bell(VteTerminal *terminal, gboolean is_audible);
 gboolean vte_terminal_get_audible_bell(VteTerminal *terminal);
 void vte_terminal_set_visible_bell(VteTerminal *terminal, gboolean is_visible);
@@ -187,6 +253,10 @@ void vte_terminal_set_color_foreground(VteTerminal *terminal,
 				       const GdkColor *foreground);
 void vte_terminal_set_color_background(VteTerminal *terminal,
 				       const GdkColor *background);
+void vte_terminal_set_color_cursor(VteTerminal *terminal,
+				   const GdkColor *cursor_background);
+void vte_terminal_set_color_highlight(VteTerminal *terminal,
+				      const GdkColor *highlight_background);
 void vte_terminal_set_colors(VteTerminal *terminal,
 			     const GdkColor *foreground,
 			     const GdkColor *background,
@@ -204,6 +274,7 @@ void vte_terminal_set_background_saturation(VteTerminal *terminal,
 					    double saturation);
 void vte_terminal_set_background_transparent(VteTerminal *terminal,
 					     gboolean transparent);
+void vte_terminal_set_opacity(VteTerminal *terminal, guint16 opacity);
 
 /* Set whether or not the cursor blinks. */
 void vte_terminal_set_cursor_blinks(VteTerminal *terminal, gboolean blink);
@@ -218,7 +289,13 @@ void vte_terminal_im_append_menuitems(VteTerminal *terminal,
 /* Set or retrieve the current font. */
 void vte_terminal_set_font(VteTerminal *terminal,
 			   const PangoFontDescription *font_desc);
+void vte_terminal_set_font_full(VteTerminal *terminal,
+				const PangoFontDescription *font_desc,
+				VteTerminalAntiAlias antialias);
 void vte_terminal_set_font_from_string(VteTerminal *terminal, const char *name);
+void vte_terminal_set_font_from_string_full(VteTerminal *terminal,
+					    const char *name,
+					    VteTerminalAntiAlias antialias);
 const PangoFontDescription *vte_terminal_get_font(VteTerminal *terminal);
 gboolean vte_terminal_get_using_xft(VteTerminal *terminal);
 void vte_terminal_set_allow_bold(VteTerminal *terminal, gboolean allow_bold);
@@ -260,6 +337,13 @@ char *vte_terminal_get_text(VteTerminal *terminal,
 						   gpointer data),
 			    gpointer data,
 			    GArray *attributes);
+char *vte_terminal_get_text_include_trailing_spaces(VteTerminal *terminal,
+						    gboolean(*is_selected)(VteTerminal *terminal,
+									   glong column,
+									   glong row,
+									   gpointer data),
+						    gpointer data,
+						    GArray *attributes);
 char *vte_terminal_get_text_range(VteTerminal *terminal,
 				  glong start_row, glong start_col,
 				  glong end_row, glong end_col,
@@ -271,7 +355,6 @@ char *vte_terminal_get_text_range(VteTerminal *terminal,
 				  GArray *attributes);
 void vte_terminal_get_cursor_position(VteTerminal *terminal,
 				      glong *column, glong *row);
-
 /* Display string matching:  clear all matching expressions. */
 void vte_terminal_match_clear_all(VteTerminal *terminal);
 
@@ -281,6 +364,8 @@ int vte_terminal_match_add(VteTerminal *terminal, const char *match);
 /* Set the cursor to be used when the pointer is over a given match. */
 void vte_terminal_match_set_cursor(VteTerminal *terminal, int tag,
 				   GdkCursor *cursor);
+void vte_terminal_match_set_cursor_type(VteTerminal *terminal,
+					int tag, GdkCursorType cursor_type);
 /* Remove a matching expression by tag. */
 void vte_terminal_match_remove(VteTerminal *terminal, int tag);
 
@@ -294,6 +379,7 @@ char *vte_terminal_match_check(VteTerminal *terminal,
 /* Set the emulation type.  Most of the time you won't need this. */
 void vte_terminal_set_emulation(VteTerminal *terminal, const char *emulation);
 const char *vte_terminal_get_emulation(VteTerminal *terminal);
+const char *vte_terminal_get_default_emulation(VteTerminal *terminal);
 
 /* Set the character encoding.  Most of the time you won't need this. */
 void vte_terminal_set_encoding(VteTerminal *terminal, const char *codeset);
@@ -305,6 +391,10 @@ const char *vte_terminal_get_status_line(VteTerminal *terminal);
 /* Get the padding the widget is using. */
 void vte_terminal_get_padding(VteTerminal *terminal, int *xpad, int *ypad);
 
+/* Attach an existing PTY master side to the terminal widget.  Use
+ * instead of vte_terminal_fork_command(). */
+void vte_terminal_set_pty(VteTerminal *terminal, int pty_master);
+
 /* Accessors for bindings. */
 GtkAdjustment *vte_terminal_get_adjustment(VteTerminal *terminal);
 glong vte_terminal_get_char_width(VteTerminal *terminal);
@@ -316,9 +406,6 @@ glong vte_terminal_get_column_count(VteTerminal *terminal);
 const char *vte_terminal_get_window_title(VteTerminal *terminal);
 const char *vte_terminal_get_icon_title(VteTerminal *terminal);
 
-/* You never saw this. */
-void _vte_terminal_accessible_ref(VteTerminal *terminal);
-
 G_END_DECLS
 
 #endif
